"""
Basic use:
    download(year)
    build(year)
    read_dataframe(year)

Helper:
    inspect(year)
"""

from boo.year import make_url
from boo.path import RawFile
from boo.curl import curl
from boo.columns import SHORT_COLUMNS, SHORT_INDEX
from boo.dataframe.canonic import canonic_df


def conditional_delete(path, force: bool):
    """Delete an exisiting file at *path* if *force* flag is set to True"""
    if force is True and path.exists():        
        path.unlink()
        print("Deleted", path)


def force_message(year, verb):
    return f"Use {verb}({year}, force=True) to overwrite existing file."


def download(year: int, force=False, directory=None):
    """Download file from Rosstat web site."""
    raw_file = RawFile(year, directory)
    path = raw_file.path
    url = make_url(year)
    conditional_delete(path, force)
    if not path.exists():
        print(f"Downloading source file for {year} from", url)
        curl(path, url)
        print("Saved as", raw_file)
    else:
        print("Already downloaded:", raw_file)
        print(force_message(year, "download"))
    return path


def read_intermediate_df(year: int, directory=None, **kwargs):
    import pandas as pd
    src = RawFile(year, directory).path
    return pd.read_csv(src, 
                       encoding="windows-1251", 
                       sep=";", 
                       header=None, 
                       usecols=SHORT_INDEX,     # read only a subset of columns
                       names=SHORT_COLUMNS.all, # give new names to columns
                       dtype=SHORT_COLUMNS.dtypes, # assert string types 
                       **kwargs)


def read_dataframe(year: int, directory=None, **kwargs):
    """Read canonic data for *year* as dataframe.

    Returns:
        pandas.DataFrame
    """
    return canonic_df(read_intermediate_df(year, directory, **kwargs))


def inspect(year: int, directory=None):
    """Diagnose local data file for *year*."""    
    raw = RawFile(year, directory)
    if raw.exists():
        print(f"      Raw CSV file: {raw}")
        if raw.mb() < 1:
            print(
                "WARNING: file size too small. "
                "Usually file size is larger than 500Mb."
            )
    else:
        raw.print_error()
    return  str(raw.path)


import pytest

from pathlib import Path


@pytest.fixture
def tempdir(tmpdir):
    return Path(tmpdir)


def test_dir(tempdir):
    assert tempdir.exists()


if __name__ == "__main__":
    pytest.main([__file__])



import pytest
from shutil import copyfile
from pathlib import Path
from tempfile import TemporaryDirectory

DATA_FOLDER = Path(__file__).parent / "data"


def raw_content():
    return (DATA_FOLDER / "raw0.csv").read_text(encoding="cp1251")


def processed_content():
    return (DATA_FOLDER / "0.csv").read_text(encoding="utf-8")


def size(path):
    return path.stat().st_size


def copy(filename, destination_folder):
    src = DATA_FOLDER / filename
    dst = Path(destination_folder) / filename
    copyfile(src, dst)


class Test_filled_directory(object):
    def setup_method(self):
        self.temp_dir = TemporaryDirectory()
        copy("raw0.csv", self.temp_dir.name)
        copy("0.csv", self.temp_dir.name)
        self.kwargs = dict(year=0, directory=self.temp_dir)

    def teardown_method(self):
        self.temp_dir.cleanup()


@pytest.fixture
def SIZE_RAW():
    return size(DATA_FOLDER / "raw0.csv")


@pytest.fixture
def SIZE_PROCESSED():
    return size(DATA_FOLDER / "0.csv")


@pytest.fixture
def directory_with_files(tmpdir):
    copy("raw0.csv", tmpdir)
    copy("0.csv", tmpdir)
    return str(tmpdir)


@pytest.fixture
def ARGS_YEAR_0(directory_with_files):
    return 0, str(directory_with_files)



"""Code descriptions based on
   http://www.consultant.ru/document/cons_doc_LAW_103394/c8c663513ad32e5a0eb8ca96753ea3e0911415db/
"""

ACCOUNT_NAMES_TEXT = """БУХГАЛТЕРСКИЙ БАЛАНС	1000
Итого внеоборотных активов	1100
Нематериальные активы	1110
Результаты исследований и разработок	1120
Нематериальные поисковые активы	1130
Материальные поисковые активы	1140
Основные средства	1150
Доходные вложения в материальные ценности	1160
Финансовые вложения	1170
Отложенные налоговые активы	1180
Прочие внеоборотные активы	1190
Итого оборотных активов	1200
Запасы	1210
Налог на добавленную стоимость по приобретенным ценностям	1220
Дебиторская задолженность	1230
Финансовые вложения (за исключением денежных эквивалентов)	1240
Денежные средства и денежные эквиваленты	1250
Прочие оборотные активы	1260
БАЛАНС (актив)	1600
Итого капитал	1300
Уставный капитал (складочный капитал, уставный фонд, вклады товарищей)	1310
Собственные акции, выкупленные у акционеров	1320
Переоценка внеоборотных активов	1340
Добавочный капитал (без переоценки)	1350
Резервный капитал	1360
Нераспределенная прибыль (непокрытый убыток)	1370
Долгосрочные заемные средства	1410
Отложенные налоговые обязательства	1420
Оценочные обязательства	1430
Прочие долгосрочные обязательства	1450
Итого долгосрочных обязательств	1400
Краткосрочные заемные обязательства	1510
Краткосрочная кредиторская задолженность	1520
Доходы будущих периодов	1530
Оценочные обязательства	1540
Прочие краткосрочные обязательства	1550
Итого краткосрочных обязательств	1500
БАЛАНС (пассив)	1700
ОТЧЕТ О ФИНАНСОВЫХ РЕЗУЛЬТАТАХ	2000
Выручка	2110
Себестоимость продаж	2120
Валовая прибыль (убыток)	2100
Коммерческие расходы	2210
Управленческие расходы	2220
Прибыль (убыток) от продаж	2200
Доходы от участия в других организациях	2310
Проценты к получению	2320
Проценты к уплате	2330
Прочие доходы	2340
Прочие расходы	2350
Прибыль (убыток) до налогообложения	2300
Текущий налог на прибыль	2410
Постоянные налоговые обязательства (активы)	2421
Изменение отложенных налоговых обязательств	2430
Изменение отложенных налоговых активов	2450
Прочее	2460
Чистая прибыль (убыток)	2400
Результат от переоценки внеоборотных активов, не включаемый в чистую прибыль (убыток)	2510
Результат от прочих операций, не включаемый в чистую прибыль (убыток) периода	2520
Совокупный финансовый результат периода	2500
Базовая прибыль (убыток) на акцию	2900
Разводненная прибыль (убыток) на акцию	2910
ОТЧЕТ О ДВИЖЕНИИ ДЕНЕЖНЫХ СРЕДСТВ	4000
Поступления - всего	4110
От продажи продукции, товаров, работ и услуг	4111
Арендных платежей, лицензионных платежей, роялти, комиссионных и иных аналогичных платежей	4112
От перепродажи финансовых вложений	4113
Прочие поступления	4119
Платежи - всего	4120
Поставщикам (подрядчикам) за сырье, материалы, работы, услуги	4121
В связи с оплатой труда работников	4122
Проценты по долговым обязательствам	4123
Налога на прибыль организаций	4124
Прочие платежи	4129
Сальдо денежных потоков от текущих операций	4100
Поступления - всего	4210
От продажи внеоборотных активов (кроме финансовых вложений)	4211
От продажи акций других организаций (долей участия)	4212
От возврата предоставленных займов, от продажи долговых ценных бумаг (прав требования денежных средств к другим лицам)	4213
Дивидендов, процентов по долговым финансовым вложениям и аналогичных поступлений от долевого участия в других организациях	4214
Прочие поступления	4219
Платежи - всего	4220
В связи с приобретением, созданием, модернизацией, реконструкцией и подготовкой к использованию внеоборотны активов	4221
В связи с приобретением акций других организаций (долей участия)	4222
В связи с приобретением долговых ценных бумаг (прав требования денежных средств к другим лицам), предоставление займов другим лицам	4223
Процентов по долговым обязательствам, включаемым в стоимость инвестиционного актива	4224
Прочие платежи	4229
Сальдо денежных потоков от инвестиционных операций	4200
Поступления - всего	4310
Получение кредитов и займов	4311
Денежных вкладов собственников (участников)	4312
От выпуска акций, увеличения долей участия	4313
От выпуска облигаций, векселей и других долговых ценных бумаг и др.	4314
Прочие поступления	4319
Платежи - всего	4320
Собственникам (участникам) в связи с выкупом у них акций (долей участия) организации или их выходом из состава участников	4321
На уплату дивидендов и иных платежей по распределению прибыли в пользу собственников (участников)	4322
В связи с погашением (выкупом) векселей и других долговы ценных бумаг, возврат кредитов и займов	4323
Прочие платежи	4329
Сальдо денежных потоков от финансовых операций	4300
Сальдо денежных потоков за отчетный период	4400
Остаток денежных средств и денежных эквивалентов на начало отчетного периода	4450
Остаток денежных средств и денежных эквивалентов на конец отчетного периода	4500
Величина влияния изменений курса иностранной валюты по отношению к рублю	4490
Остаток средств на начало отчетного года	6100
Поступило средств - всего	6200
Вступительные взносы	6210
Членские взносы	6215
Целевые взносы	6220
Добровольные имущественные взносы и пожертвования	6230
Прибыль от предпринимательской деятельности организации	6240
Прочие	6250
Использовано средств - всего	6300
Расходы на целевые мероприятия	6310
социальная и благотворительная помощь	6311
проведение конференций, совещаний, семинаров и т.п.	6312
иные мероприятия	6313
Расходы на содержание аппарата управления	6320
расходы, связанные с оплатой труда (включая начисления)	6321
выплаты, связанные с оплатой труда	6322
расходы на служебные командировки и деловые поездки	6323
содержание помещений, зданий, автомобильного транспорта и иного имущества (кроме ремонта)	6324
ремонт основных средств и иного имущества	6325
прочие	6326
Приобретение основных средств, инвентаря и иного имущества	6330
Прочие	6350
Остаток средств на конец отчетного года	6400"""


def items_from_doc(doc: str):
    for x in doc.split("\n"):
        y = x.split("\t")
        try:
            yield y[1], y[0]
        except IndexError:
            raise ValueError(y)


def code_to_description(code: str, doc=ACCOUNT_NAMES_TEXT):
    """Return account text description by code."""
    acc_names = dict(items_from_doc(doc))
    return acc_names.get(code)


def okv(text):
    return f"Код ОКВЭД {text} уровня"


ADDITIONAL_DICT = {
    "ok1": okv("первого"),
    "ok2": okv("второго"),
    "ok3": okv("третьего"),
    "org": "Тип юридического лица (часть наименования организации)",
    "title": "Короткое название организации",
    "region": "Код региона по ИНН",
    "p": "Прибыль (убыток) до налогообложения",
}


def whatis(varname: str, additional_codes: dict = ADDITIONAL_DICT):
    """Return text description for *varname*."""
    from boo.columns import unlag, varname_to_code
    varname = unlag(varname)
    try:
        return additional_codes[varname]
    except KeyError:
        return code_to_description(code=varname_to_code(varname))

    


from boo.columns import SHORT_COLUMNS, TTL_COLUMNS
from boo.columns import varname_to_code, code_to_varname


def test_varname_to_code_works_for_one_arg():
    assert varname_to_code("ta") == "1600"

def test_code_to_varname_works_for_one_arg():
    assert code_to_varname('4320') == "cf_fin_out"


def test_sc_text():
    assert SHORT_COLUMNS.text == [
        "name",
        "okpo",
        "okopf",
        "okfs",
        "okved",
        "inn",
        "unit",
        "report_type",
        "date_published",
    ]


from boo.whatis import whatis


def test_whatis_returns_name():
    assert whatis("cash") == "Денежные средства и денежные эквиваленты"


def test_whatis_works_with_lag():
    assert whatis("ta") == whatis("ta_lag")


def test_whatis_returns_additional_name():
    assert whatis("ok1") == "Код ОКВЭД первого уровня"


def test_whatis_returns_on_shorthand():
    assert whatis("p") == "Прибыль (убыток) до налогообложения"


def test_whatis_is_confused():
    assert whatis("aha!") is None



class UnclassifiableCodeError(ValueError):
    pass


def commas(xs):
    return ", ".join(map(str, xs))


class WrongYearError(ValueError):
    def __init__(self, year: int, allowed: [int]):
        super().__init__(
            "Year not supported.\n"
            f"Provided year: {year}.\n"
            f"Allowed years: {commas(allowed)}.\n"
            "Update timestamp definitions in year.py if necessary."
        )


class DirectoryNotFound(FileNotFoundError):
    pass


class NoRawFileError(FileNotFoundError):
    def __init__(self, year):
        super().__init__(
            f"Raw CSV file not downloaded for {year}. " f"Try `boo.download({year})`."
        )



from pathlib import Path

from boo.errors import DirectoryNotFound, NoRawFileError
from boo.helper import as_mb


def default_data_folder() -> Path:
    home = Path.home() / ".boo"
    home.mkdir(exist_ok=True)
    return home


def get_folder(directory=None) -> Path:
    if directory is None:
        return default_data_folder()
    elif Path(directory).is_dir():
        return Path(directory)
    else:
        raise DirectoryNotFound(directory)


def file(year, tag="", directory=None):
    return get_folder(directory) / f"{tag}{year}.csv"


class RawFile:
    encoding = "cp1251"
    
    def __init__(self, year, directory):
        self.path = file(year, "raw", directory)
        self.year = year

    def content(self):
        return self.path.read_text(encoding="cp1251")

    def size(self):
        return self.path.stat().st_size

    def mb(self):
        return as_mb(self.size())

    def folder(self):
        return str(self.path.parent)

    def __str__(self):
        prefix = str(self.path)
        try:
            return prefix + f" ({self.mb()}M)"
        except FileNotFoundError:
            return prefix + " (does not exist)"

    def __repr__(self):
        return repr(self.path)

    def exists(self):
        return self.path.exists()

    def assert_exists(self):
        if not self.exists():
            raise NoRawFileError(self.year)
            
    def print_error(self):
        try:
            self.assert_exists()
        except FileNotFoundError as e:
            print(e)
            


import numpy
import pandas as pd
from boo.columns import SHORT_COLUMNS
from boo.errors import UnclassifiableCodeError


QUOTE_CHAR = '"'
EMPTY = int(0)
NUMERIC_COLUMNS = SHORT_COLUMNS.numeric


def adjust_rub(df, cols=NUMERIC_COLUMNS):
    rows = df.unit == "385"
    df.loc[rows, cols] = df.loc[rows, cols].multiply(1000)
    df.loc[rows, "unit"] = "384"
    rows = df.unit == "383"
    df.loc[rows, cols] = df.loc[rows, cols].divide(1000).round(0).astype(int)
    df.loc[rows, "unit"] = "384"
    return df


def dequote(name: str):
    """Split company *name* to organisation and title."""
    # Warning: will not work well on company names with more than 4 quotechars
    parts = name.split(QUOTE_CHAR)
    org = parts[0].strip()
    cnt = name.count(QUOTE_CHAR)
    if cnt == 2:
        title = parts[1].strip()
    elif cnt > 2:
        title = QUOTE_CHAR.join(parts[1:])
    else:
        title = name
    return org, title.strip()


def replace_names(title: str):
    return (
        title.replace("ПУБЛИЧНОЕ АКЦИОНЕРНОЕ ОБЩЕСТВО", "ПАО")
        .replace("ОТКРЫТОЕ АКЦИОНЕРНОЕ ОБЩЕСТВО", "ОАО")
        .replace("АКЦИОНЕРНОЕ ОБЩЕСТВО ЭНЕРГЕТИКИ И ЭЛЕКТРИФИКАЦИИ", "AO энерго")
        .replace("НЕФТЕПЕРЕРАБАТЫВАЮЩИЙ ЗАВОД", "НПЗ")
        .replace("ГЕНЕРИРУЮЩАЯ КОМПАНИЯ ОПТОВОГО РЫНКА ЭЛЕКТРОЭНЕРГИИ", "ОГК")
        .replace("ГОРНО-ОБОГАТИТЕЛЬНЫЙ КОМБИНАТ", "ГОК")
    )


def add_title(df):
    s_ = df.name.apply(dequote)
    df["org"] = s_.apply(lambda x: x[0])
    df["title"] = s_.apply(lambda x: replace_names(x[1]))
    return df


def rename_rows(df):
    RENAME_DICT = {
        "2460066195": "РусГидро",
        "4716016979": "ФСК ЕЭС",
        "7702038150": "Московский метрополитен",
        "7721632827": "Концерн Росэнергоатом",
        "7706664260": "Атомэнергопром",
        "7703683145": "Холдинг ВТБ Капитал АЙ БИ",
        "9102048801": "Черноморнефтегаз",
        "7736036626": "РИТЭК",
    }
    keys = RENAME_DICT.keys()
    ix = df.index.isin(keys)
    if not ix.any():
        return df
    sub = df.loc[ix, "title"]
    for k, v in RENAME_DICT.items():
        if k in sub.index:
            sub.loc[k] = v
    df.loc[ix, "title"] = sub
    return df


def split_okved(code_string: str):
    """Get 3 levels of OKVED codes from *code_string*."""
    if code_string.count(".") > 2:
        raise UnclassifiableCodeError(code_string)
    try:
        codes = [int(x) for x in code_string.split(".")]
    except ValueError:
        raise UnclassifiableCodeError(code_string)
    return codes + [0] * (3 - len(codes))


def add_okved_subcode(df):
    df["ok1"], df["ok2"], df["ok3"] = zip(*df.okved.apply(split_okved))
    return df


def fst(x):
    try:
        return int(x[0:2])
    except TypeError:
        return 0


def add_region(df):
    df["region"] = df.inn.apply(fst)
    return df


def more_columns(df):
    return add_okved_subcode(add_region(add_title(df)))


def canonic_df(df):
    """Преобразовать данные внтури датафрейма:

    - Привести все строки к одинаковым единицам измерения (тыс. руб.)
    - Убрать  неиспользуемые колонки (date_revised, report_type)
    - Новые колонки:
        * короткое название компании
        * три уровня кода ОКВЭД
        * регион (по ИНН)

    """
    df_ = add_okved_subcode(add_region(add_title(df)))
    df_ = adjust_rub(df_)
    df_ = df_.set_index("inn")
    df_ = rename_rows(df_)
    return df_[canonic_columns()]


def canonic_columns(numeric=SHORT_COLUMNS.numeric):
    return (
        ["title", "org", "okpo", "okopf", "okfs", "okved"]
        + ["unit"]
        + ["ok1", "ok2", "ok3", "region"]
        + numeric
    )


def is_numeric_column(name, numeric=SHORT_COLUMNS.numeric):
    return name in numeric


def columns_typed_as_integer(numeric=SHORT_COLUMNS.numeric):
    return numeric + ["ok1", "ok2", "ok3", "region"]

